#include <stdio.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#include "tsc.h"

struct list {
        struct list * next;
        unsigned int data;
};

static inline void prefetch(const void *x)
{
    __asm__ __volatile__("prefetchnta (%0)"
                          : : "r" (x));
}

void allocate_list(struct list *head, int c, int p)
{
    struct list *x, *y;
    int i;

    head->next = y = malloc(sizeof(struct list)+p);
    if (!y) {
        fprintf(stderr, "Memory allocation error\n");
        exit(1);
    }

    y->data = 0;

    for (i=1; i<c; i++) {
        x = malloc(sizeof(struct list)+p);
        if (!x) {
            fprintf(stderr, "Memory allocation error\n");
            exit(1);
        }
        y->next = x;
        x->data = i;
        y = x;
    }
    y->next = head;
}

void release_list(struct list *head)
{
    struct list *x = head->next, *y;
    do {
        y = x->next;
	free(x);
        x = y;
    } while (x != head);
}

void process_list_element(int *data, int *d, int delay)
{
	int l;
	*d = *data;
	*data = (*d)*(*d);
        for (l=0;l<delay*100;++l) ;
}

int main(int argc, char *argv[])
{
    int c, d, s, t;
    double time_p, time_n, time_d;
    struct list Head;
    struct list *x, *head = &Head;

    if (argc < 4) {
        fprintf(stderr, "Usage: %s <el_size> <el_num> <el_delay>\n", argv[0]);
        return 1;
    }

    calibrate_tsc(0);
    
    s = atoi(argv[1]);
    c = atoi(argv[2]);
    d = atoi(argv[3]);

    if (s < 0 || c <= 0)
        return 1;
    
{
    int q, i;
    start_timing(1);
    for (i=0; i<10000; ++i)
	    process_list_element(&t, &q, d);
    stop_timing(1);
    time_d = timing_result();
    fprintf(stderr, "delay on each element: %.3f nanoseconds\n",
	time_d*1000000.0/10000);
}

    allocate_list(head, c, s);

    start_timing(1);
    for (x=head->next,t=0; prefetch(x->next), x != head; x = x->next)
	process_list_element(&x->data, &t, d);
    stop_timing(1);
    time_p = timing_result();

    release_list(head);
    usleep(1000);
    allocate_list(head, c, s);
    
    start_timing(1);
    for (x=head->next,t=0; x != head; x = x->next)
	process_list_element(&x->data, &t, d);
    stop_timing(1);
    time_n = timing_result();
    
    fprintf(stderr, "n=%d, s=%d, w/ prefetch=%.3f msec, w/o prefetch=%.3f msec (%.3f)\n", 
            c, s, time_p, time_n, (time_p-time_n)*1000000.0/c); 
 
    printf("%d %d %f %f %f\n", c, s, time_d, time_p, time_n);
#if 0
    start_timing(1);
    for (d=0; d<c; ++d)
        prefetch(head->next);
    stop_timing(1);
    time_p = timing_result();
    fprintf(stderr, "Prefetch time: %.3f msec, %.3f nsec/prefetch\n",
    	time_p, time_p*1000000.0/c);
#endif
    return 0;
}

